"""
API client wrapper for Anthropic's Claude API.
"""
from typing import Dict, List, Optional, Any
import anthropic
import json

class ApiClient:
    """
    Wrapper for the Anthropic API client.
    """
    
    def __init__(self, api_key: Optional[str] = None, debug_mode: bool = False):
        """
        Initialize the API client.
        
        Args:
            api_key: Anthropic API key. If None, will try to get from environment variable.
            debug_mode: If True, print debug information about API calls.
        """
        self.client = anthropic.Anthropic(api_key=api_key)
        self.debug_mode = debug_mode
    
    def create_message(self, 
                      model: str,
                      messages: List[Dict[str, Any]],
                      max_tokens: int = 1024,
                      temperature: float = 0.7,
                      system: Optional[str] = None,
                      tools: Optional[List[Dict]] = None,
                      tool_choice: Optional[Dict] = None) -> anthropic.types.Message:
        """
        Create a message using the Anthropic API.
        
        Args:
            model: Claude model to use
            messages: List of messages in the conversation
            max_tokens: Maximum number of tokens to generate
            temperature: Temperature for generation
            system: System prompt to guide the model's behavior
            tools: List of tool schemas
            tool_choice: Tool choice configuration
            
        Returns:
            Anthropic API response
        """
        # Prepare API call parameters
        api_params = {
            "model": model,
            "max_tokens": max_tokens,
            "temperature": temperature,
            "messages": messages
        }
        
        # Add system prompt if provided
        if system:
            api_params["system"] = system
        
        # Add tools if provided
        if tools:
            api_params["tools"] = tools
            
            # Add tool_choice if provided
            if tool_choice:
                api_params["tool_choice"] = tool_choice
        
        # Debug mode: print API call parameters
        if self.debug_mode:
            print("\n===== DEBUG: API REQUEST =====")
            debug_params = api_params.copy()
            # Format messages for better readability
            if "messages" in debug_params:
                print(f"Messages ({len(debug_params['messages'])} items):")
                for i, msg in enumerate(debug_params["messages"]):
                    print(f"  Message {i+1}:")
                    for k, v in msg.items():
                        if k == "content" and isinstance(v, list):
                            print(f"    {k}: [")
                            for item in v:
                                print(f"      {json.dumps(item, indent=2)}")
                            print("    ]")
                        else:
                            print(f"    {k}: {v}")
            
            # Format tools for better readability
            if "tools" in debug_params:
                print(f"\nTools ({len(debug_params['tools'])} items):")
                for i, tool in enumerate(debug_params["tools"]):
                    print(f"  Tool {i+1}: {json.dumps(tool, indent=2)}")
            
            # Print other parameters
            print("\nOther parameters:")
            for k, v in debug_params.items():
                if k not in ["messages", "tools"]:
                    print(f"  {k}: {v}")
            print("===== END DEBUG =====\n")
        
        # Make the API call
        return self.client.messages.create(**api_params)



#!/usr/bin/env python3
from claudine import Agent
import sys

def main():
    """
    Example that demonstrates how to use the built-in token tracking and cost functionality.
    """
    # Initialize Agent
    agent = Agent(max_tokens=1000)
    
    # First prompt
    first_prompt = "Write a short poem about programming."
    first_response = agent.process_prompt(first_prompt)
    
    print(f"User: {first_prompt}")
    print(f"Claude: {first_response}")
    
    # Get token usage for the first message
    token_info = agent.get_token_usage()
    
    print("\nToken Usage Information:")
    print(f"Input tokens: {token_info.text_usage.input_tokens}")
    print(f"Output tokens: {token_info.text_usage.output_tokens}")
    print(f"Total tokens: {token_info.text_usage.total_tokens}")
    
    # Get cost information
    cost_info = agent.get_cost()
    
    print("\nCost Information:")
    print(f"Input cost: {cost_info.format_input_cost()} {cost_info.unit}")
    print(f"Output cost: {cost_info.format_output_cost()} {cost_info.unit}")
    print(f"Total cost: {cost_info.format_total_cost()} {cost_info.unit}")
    
    # Second prompt
    second_prompt = "Explain how token counting works in large language models."
    second_response = agent.process_prompt(second_prompt)
    
    print(f"\nUser: {second_prompt}")
    print(f"Claude: {second_response}")
    
    # Get updated token usage
    token_info = agent.get_token_usage()
    
    print("\nUpdated Token Usage:")
    print(f"Total input tokens: {token_info.total_usage.input_tokens}")
    print(f"Total output tokens: {token_info.total_usage.output_tokens}")
    print(f"Total tokens: {token_info.total_usage.total_tokens}")
    
    # Get updated cost information
    cost_info = agent.get_cost()
    
    print("\nUpdated Cost Information:")
    print(f"Total cost: {cost_info.format_total_cost()} {cost_info.unit}")
    
    return 0

if __name__ == "__main__":
    sys.exit(main())



import inspect
from typing import Any, Callable, Dict, List, Optional, Union, get_type_hints

class ToolManager:
    """
    Manages tool registration, schema generation, and execution for the Agent.
    """
    
    def __init__(self):
        """Initialize an empty tool manager."""
        self.registered_tools = {}
        
        # Default callbacks
        self.pre_tool_callback = self._default_pre_tool_callback
        self.post_tool_callback = self._default_post_tool_callback
    
    def _default_pre_tool_callback(self, tool_name: str, tool_input: Dict[str, Any], 
                                    preamble_text: str) -> Dict[str, Any]:
        """
        Default callback called before a tool is executed.
        
        Args:
            tool_name: Name of the tool being called
            tool_input: Input parameters for the tool
            preamble_text: Any text Claude generated before the tool call
            
        Returns:
            Possibly modified tool input
        """
        return tool_input
    
    def _default_post_tool_callback(self, tool_name: str, tool_input: Dict[str, Any], 
                                     result: Any) -> Any:
        """
        Default callback called after a tool is executed.
        
        Args:
            tool_name: Name of the tool that was called
            tool_input: Input parameters that were passed to the tool
            result: Result returned by the tool
            
        Returns:
            Possibly modified result
        """
        return result
    
    def set_tool_callbacks(self, pre_callback: Optional[Callable] = None, 
                            post_callback: Optional[Callable] = None):
        """
        Set custom callbacks for tool calls.
        
        Args:
            pre_callback: Function to call before executing a tool
            post_callback: Function to call after executing a tool
        """
        if pre_callback:
            self.pre_tool_callback = pre_callback
        if post_callback:
            self.post_tool_callback = post_callback
    
    def register_tool(self, func: Callable, name: Optional[str] = None, description: Optional[str] = None):
        """
        Register a function as a tool.
        
        Args:
            func: The function to register
            name: Optional custom name for the tool (defaults to function name)
            description: Optional description of what the tool does
            
        Returns:
            The original function
        """
        func_name = name or func.__name__
        func_description = description or func.__doc__ or f"Function {func_name}"
        
        # Extract clean description
        if func_description:
            func_description = func_description.split('\n\n')[0]
        
        # Build the schema based on function signature and type hints
        schema = self._build_schema_from_function(func)
        
        # Store the function and its schema
        self.registered_tools[func_name] = {
            "function": func,
            "schema": {
                "name": func_name,
                "description": func_description,
                "input_schema": schema
            }
        }
        
        return func
    
    def register_tools(self, tools: List[Callable]):
        """
        Register multiple tools at once.
        
        Args:
            tools: List of functions to register as tools
        """
        for tool in tools:
            self.register_tool(tool)
    
    def _build_schema_from_function(self, func: Callable) -> Dict:
        """
        Build a JSON schema based on the function's signature and type hints.
        
        Args:
            func: The function to analyze
            
        Returns:
            A JSON schema dictionary describing the function parameters
        """
        signature = inspect.signature(func)
        hints = get_type_hints(func)
        
        properties = {}
        required = []
        
        for param_name, param in signature.parameters.items():
            # Skip self parameter for methods
            if param_name == 'self':
                continue
                
            param_type = hints.get(param_name, Any)
            param_doc = self._extract_param_doc(func, param_name)
            
            # Handle various types
            if param_type == str:
                param_schema = {"type": "string"}
            elif param_type == int:
                param_schema = {"type": "integer"}
            elif param_type == float:
                param_schema = {"type": "number"}
            elif param_type == bool:
                param_schema = {"type": "boolean"}
            elif hasattr(param_type, '__origin__') and param_type.__origin__ is list:
                param_schema = {"type": "array"}
            elif hasattr(param_type, '__origin__') and param_type.__origin__ is dict:
                param_schema = {"type": "object"}
            else:
                # Default to string for complex types
                param_schema = {"type": "string"}
            
            # Add description if available
            if param_doc:
                param_schema["description"] = param_doc
                
            properties[param_name] = param_schema
            
            # Mark as required if it doesn't have a default value
            if param.default == inspect.Parameter.empty:
                required.append(param_name)
        
        schema = {
            "type": "object",
            "properties": properties
        }
        
        if required:
            schema["required"] = required
            
        return schema
    
    def _extract_param_doc(self, func: Callable, param_name: str) -> Optional[str]:
        """
        Extract parameter documentation from function docstring.
        
        Args:
            func: The function to analyze
            param_name: The parameter name to find documentation for
            
        Returns:
            Parameter description if found, None otherwise
        """
        if not func.__doc__:
            return None
            
        docstring = inspect.cleandoc(func.__doc__)
        
        # Try to find param in the docstring using common formats
        # Look for :param param_name: or Args: param_name:
        param_patterns = [
            f":param {param_name}:", 
            f"@param {param_name}:",
            f"Args:\n    {param_name}:",
            f"Arguments:\n    {param_name}:"
        ]
        
        for pattern in param_patterns:
            if pattern in docstring:
                parts = docstring.split(pattern, 1)[1].strip()
                desc = parts.split('\n', 1)[0].strip()
                return desc
                
        return None
    
    def get_tools_schemas(self) -> List[Dict]:
        """
        Get the schema definitions for all registered tools.
        
        Returns:
            List of tool schema definitions
        """
        return [tool_info["schema"] for tool_info in self.registered_tools.values()]
    
    def get_tools_names(self) -> List[str]:
        """
        Get the names of all registered tools.
        
        Returns:
            List of tool names
        """
        return list(self.registered_tools.keys())
    
    def execute_tool(self, tool_name: str, tool_input: Dict[str, Any], preamble_text: str) -> Any:
        """
        Execute a registered tool with the given input.
        
        Args:
            tool_name: Name of the tool to execute
            tool_input: Input parameters for the tool
            preamble_text: Any text Claude generated before the tool call
            
        Returns:
            Result of the tool execution
            
        Raises:
            ValueError: If the tool is not registered
        """
        if tool_name not in self.registered_tools:
            raise ValueError(f"Tool '{tool_name}' is not registered")
            
        tool_info = self.registered_tools[tool_name]
        func = tool_info["function"]
        
        # Apply pre-execution callback
        modified_input = self.pre_tool_callback(tool_name, tool_input, preamble_text)
        
        # Execute the tool
        error = None
        result = None
        try:
            result = func(**modified_input)
        except Exception as e:
            error = e
            
        # Apply post-execution callback
        return self.post_tool_callback(tool_name, modified_input, result)



"""
Token tracking functionality for Claude.
Provides tools for tracking token usage and calculating costs across
conversations, including separate tracking for text and tool interactions.
"""
from typing import Dict, List, Optional, Union
from .api.models import TokenUsage, TokenUsageInfo, ModelPricing, TokenPricing, TokenCost

# Default model to use for pricing
DEFAULT_MODEL = "claude-3-7-sonnet-20250219"

# Model pricing data
MODEL_PRICING = {
    DEFAULT_MODEL: ModelPricing(
        input_tokens=TokenPricing(
            cost_per_million_tokens=3.0,
            unit="USD"
        ),
        output_tokens=TokenPricing(
            cost_per_million_tokens=15.0,
            unit="USD"
        )
    )
}

class TokenTracker:
    """
    Tracks token usage for Claude API calls.
    """
    
    def __init__(self):
        """Initialize an empty token tracker."""
        self.messages = {}  # Dictionary to store message token usage by message ID
    
    def add_message(self, message_id: str, input_tokens: int, output_tokens: int, 
                   is_tool_related: bool = False, tool_name: Optional[str] = None,
                   parent_message_id: Optional[str] = None):
        """
        Add a message's token usage to the tracker.
        
        Args:
            message_id: Unique ID of the message
            input_tokens: Number of input tokens used
            output_tokens: Number of output tokens used
            is_tool_related: Whether this message is part of a tool call sequence
            tool_name: Name of the tool if is_tool_related is True
            parent_message_id: ID of the parent message that initiated the tool call
        """
        self.messages[message_id] = {
            "message_id": message_id,
            "input_tokens": input_tokens,
            "output_tokens": output_tokens,
            "is_tool_related": is_tool_related,
            "tool_name": tool_name,
            "parent_message_id": parent_message_id
        }
    
    def get_token_usage(self, message_id: Optional[str] = None) -> Union[Dict, TokenUsageInfo]:
        """
        Get token usage information for a specific message or consolidated usage.
        
        Args:
            message_id: Optional ID of the message to get token usage for.
                        If None, returns consolidated token usage information.
        
        Returns:
            If message_id is provided: Token usage information for that message
            If message_id is None: TokenUsageInfo with text and tool usage information
        """
        if message_id:
            return self.messages.get(message_id, {})
        
        # Calculate total usage
        total_input = sum(msg["input_tokens"] for msg in self.messages.values())
        total_output = sum(msg["output_tokens"] for msg in self.messages.values())
        
        # Calculate tool-related usage
        tool_messages = [msg for msg in self.messages.values() if msg["is_tool_related"]]
        tool_input = sum(msg["input_tokens"] for msg in tool_messages)
        tool_output = sum(msg["output_tokens"] for msg in tool_messages)
        
        # Calculate text-only usage
        text_input = total_input - tool_input
        text_output = total_output - tool_output
        
        # Create TokenUsage objects
        text_usage = TokenUsage(
            input_tokens=text_input,
            output_tokens=text_output
        )
        
        tools_usage = TokenUsage(
            input_tokens=tool_input,
            output_tokens=tool_output
        )
        
        # Calculate usage by tool
        by_tool = {}
        for msg in self.messages.values():
            if msg["is_tool_related"] and msg["tool_name"]:
                tool_name = msg["tool_name"]
                
                if tool_name not in by_tool:
                    by_tool[tool_name] = {
                        "input_tokens": 0,
                        "output_tokens": 0
                    }
                
                by_tool[tool_name]["input_tokens"] += msg["input_tokens"]
                by_tool[tool_name]["output_tokens"] += msg["output_tokens"]
        
        # Convert by_tool to use TokenUsage objects
        by_tool_usage = {}
        for tool_name, usage in by_tool.items():
            by_tool_usage[tool_name] = TokenUsage(
                input_tokens=usage["input_tokens"],
                output_tokens=usage["output_tokens"]
            )
        
        return TokenUsageInfo(
            text_usage=text_usage,
            tools_usage=tools_usage,
            by_tool=by_tool_usage
        )
    
    def get_cost(self, message_id: Optional[str] = None) -> TokenCost:
        """
        Get cost information for token usage.
        
        Args:
            message_id: Optional ID of the message to get cost for.
                        If None, returns consolidated cost information.
        
        Returns:
            TokenCost object with cost information
        """
        # Get token usage
        usage = self.get_token_usage(message_id)
        
        # Get pricing for the default model
        pricing = MODEL_PRICING.get(DEFAULT_MODEL)
        if not pricing:
            return TokenCost(
                input_cost=0.0,
                output_cost=0.0,
                unit="USD"
            )
        
        # If it's a single message, calculate cost for that message
        if isinstance(usage, dict):
            input_cost = pricing.input_tokens.calculate_cost(usage.get("input_tokens", 0))
            output_cost = pricing.output_tokens.calculate_cost(usage.get("output_tokens", 0))
            
            return TokenCost(
                input_cost=input_cost,
                output_cost=output_cost,
                unit=pricing.input_tokens.unit
            )
        
        # For consolidated usage, extract the total cost from the cost dictionary
        cost_dict = usage.calculate_cost(pricing)
        return TokenCost(
            input_cost=cost_dict["total_cost"].input_cost,
            output_cost=cost_dict["total_cost"].output_cost,
            unit=cost_dict["total_cost"].unit
        )
    
    def reset(self):
        """Reset all token usage data."""
        self.messages = {}



#!/usr/bin/env python3
from claudine import Agent
import sys
import os
from datetime import datetime
import random

def main():
    """
    Example that demonstrates how the agent maintains awareness of previous tool results
    and can reference them in subsequent responses.
    """
    
    # Define tool functions
    def generate_random_number(min_val=1, max_val=100):
        """Generates a random number between min_val and max_val."""
        number = random.randint(min_val, max_val)
        return f"Generated random number: {number}"
    
    def check_if_prime(number):
        """Checks if a number is prime."""
        try:
            num = int(number)
            if num < 2:
                return f"{num} is not a prime number."
            for i in range(2, int(num**0.5) + 1):
                if num % i == 0:
                    return f"{num} is not a prime number. It is divisible by {i}."
            return f"{num} is a prime number."
        except ValueError:
            return f"Error: '{number}' is not a valid integer."
    
    def get_current_time():
        """Returns the current date and time."""
        now = datetime.now()
        return now.strftime("%Y-%m-%d %H:%M:%S")
    
    # Define tools list
    tools = [generate_random_number, check_if_prime, get_current_time]
    
    # Define instructions for the agent
    instructions = """
    You are a helpful AI assistant. Your primary goal is to assist the user with 
    information and general questions. When using tools, remember the results
    and be able to reference them in subsequent responses.
    """
    
    # Initialize Agent
    agent = Agent(instructions=instructions, tools=tools)
    
    # First prompt that uses a tool
    first_prompt = "Generate a random number for me."
    first_response = agent.process_prompt(first_prompt)
    
    print(f"User: {first_prompt}")
    print(f"Claude: {first_response}")
    
    # Second prompt that should reference the result of the first tool
    second_prompt = "Is the number you just generated a prime number?"
    second_response = agent.process_prompt(second_prompt)
    
    print(f"\nUser: {second_prompt}")
    print(f"Claude: {second_response}")
    
    # Third prompt to test memory of both previous tool results
    third_prompt = "What time is it now? And remind me what was the random number you generated and whether it was prime."
    third_response = agent.process_prompt(third_prompt)
    
    print(f"\nUser: {third_prompt}")
    print(f"Claude: {third_response}")
    
    return 0

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
from claudine import Agent
import sys
import os

def main():
    """
    Example that demonstrates how the agent maintains message history
    and can reference previous answers in subsequent responses.
    """
    # Check for API key
    if "ANTHROPIC_API_KEY" not in os.environ:
        print("Error: ANTHROPIC_API_KEY environment variable not set")
        print("Please set your Anthropic API key with:")
        print("  export ANTHROPIC_API_KEY=your_api_key_here (Linux/Mac)")
        print("  set ANTHROPIC_API_KEY=your_api_key_here (Windows)")
        return 1
    
    # Define instructions for the agent
    instructions = """
    You are a helpful AI assistant. Your primary goal is to assist the user with 
    information and general questions. When the user asks follow-up questions,
    you should remember and reference your previous answers.
    """
    
    # Initialize Agent
    agent = Agent(instructions=instructions)
    
    # First question
    first_prompt = "Generate a random 4-digit number for me to use as a PIN code."
    first_response = agent.process_prompt(first_prompt)
    
    print(f"User: {first_prompt}")
    print(f"Claude: {first_response}")
    
    # Second question that references the first answer
    second_prompt = "What was the PIN code you just gave me? Also, tell me what would happen if I added 1 to that number."
    second_response = agent.process_prompt(second_prompt)
    
    print(f"\nUser: {second_prompt}")
    print(f"Claude: {second_response}")
    
    # Third question to further test memory
    third_prompt = "Can you remind me what the original PIN was and what number you calculated when adding 1 to it?"
    third_response = agent.process_prompt(third_prompt)
    
    print(f"\nUser: {third_prompt}")
    print(f"Claude: {third_response}")
    
    return 0

if __name__ == "__main__":
    sys.exit(main())



"""
Common utility functions for Claudine.
"""
from typing import Dict, List, Optional, Any, Union, Tuple
import uuid
import json

def generate_message_id() -> str:
    """
    Generate a unique message ID.
    
    Returns:
        Unique message ID
    """
    return f"msg_{uuid.uuid4().hex[:24]}"

def extract_text_content(content_blocks: List[Any]) -> str:
    """
    Extract text content from content blocks.
    
    Args:
        content_blocks: List of content blocks from Claude's response
        
    Returns:
        Concatenated text content
    """
    text_content = ""
    
    for block in content_blocks:
        if hasattr(block, 'type') and block.type == "text":
            text_content += block.text
    
    return text_content

def format_tool_result(tool_use_id: str, result: Union[str, Tuple[str, bool]]) -> Dict:
    """
    Format a tool result for sending to Claude.
    
    Args:
        tool_use_id: ID of the tool use
        result: Result of the tool execution, either a string or a tuple of (content, is_error)
        
    Returns:
        Formatted tool result
    """
    if isinstance(result, tuple) and len(result) == 2 and isinstance(result[1], bool):
        content, is_error = result
        return {
            "type": "tool_result",
            "tool_use_id": tool_use_id,
            "content": content,
            "is_error": is_error
        }
    else:
        return {
            "type": "tool_result",
            "tool_use_id": tool_use_id,
            "content": result
        }



#!/usr/bin/env python3
from claudine import Agent
import sys

def main():
    """
    Example that demonstrates how to obtain token usage and cost information
    after making calls to the Claudine Agent.
    """
    
    # Initialize Agent
    agent = Agent(max_tokens=1000, temperature=0.7)
    
    # Simple message to the API
    response = agent.process_prompt("Write a short poem about programming.")
    
    # Print the response
    print("Claude's response:")
    print(response)
    print("\n" + "-" * 50 + "\n")
    
    # Get token usage information
    token_info = agent.get_token_usage()
    
    print("Token Usage Information:")
    print(f"Text input tokens: {token_info.text_usage.input_tokens}")
    print(f"Text output tokens: {token_info.text_usage.output_tokens}")
    print(f"Text total tokens: {token_info.text_usage.total_tokens}")
    print(f"Tool input tokens: {token_info.tools_usage.input_tokens}")
    print(f"Tool output tokens: {token_info.tools_usage.output_tokens}")
    print(f"Tool total tokens: {token_info.tools_usage.total_tokens}")
    print(f"Total input tokens: {token_info.total_usage.input_tokens}")
    print(f"Total output tokens: {token_info.total_usage.output_tokens}")
    print(f"Total tokens: {token_info.total_usage.total_tokens}")
    
    # Get cost information
    cost_info = agent.get_cost()
    
    print("\nCost Information:")
    print(f"Input cost: {cost_info.format_input_cost()} {cost_info.unit}")
    print(f"Output cost: {cost_info.format_output_cost()} {cost_info.unit}")
    print(f"Total cost: {cost_info.format_total_cost()} {cost_info.unit}")
    
    # Example with a longer prompt
    print("\n" + "-" * 50 + "\n")
    print("Example with a longer prompt:")
    
    longer_response = agent.process_prompt("Explain how token counting works in large language models and why it matters for API usage.")
    
    # Print the response
    print("Claude's response:")
    print(longer_response)
    print("\n" + "-" * 50 + "\n")
    
    # Get updated token usage information
    token_info = agent.get_token_usage()
    
    print("Updated Token Usage Information:")
    print(f"Total input tokens: {token_info.total_usage.input_tokens}")
    print(f"Total output tokens: {token_info.total_usage.output_tokens}")
    print(f"Total tokens: {token_info.total_usage.total_tokens}")
    
    # Get updated cost information
    cost_info = agent.get_cost()
    
    print("\nUpdated Cost Information:")
    print(f"Total cost: {cost_info.format_total_cost()} {cost_info.unit}")
    
    return 0

if __name__ == "__main__":
    sys.exit(main())



"""
Tool management functionality for Claude.
Provides a system for registering, managing, and executing tools with Claude,
including schema generation, callbacks, and execution handling.
"""
from typing import Dict, List, Optional, Callable, Any, Tuple, Union
import inspect
import json
from .schema import generate_tool_schema

class ToolManager:
    """
    Manages tool registration, schema generation, and execution.
    """
    
    def __init__(self):
        """Initialize an empty tool manager."""
        self.tools = {}
        self.pre_tool_callback = None
        self.post_tool_callback = None
        self.text_editor_tool = None
    
    def register_tools(self, tools: List[Callable]):
        """
        Register multiple tools.
        
        Args:
            tools: List of functions to register as tools
        """
        for tool in tools:
            # Get tool name (use function name if not provided)
            tool_name = tool.__name__
            
            # Store the function
            self.tools[tool_name] = tool
            
            # Check if this is a text editor tool
            if tool_name == "str_replace_editor":
                self.text_editor_tool = tool
    
    def set_tool_callbacks(self, pre_callback: Optional[Callable] = None, 
                             post_callback: Optional[Callable] = None):
        """
        Set callbacks for tool execution.
        
        Args:
            pre_callback: Function to call before tool execution.
                             Must have signature (tool_name: str, tool_input: Dict[str, Any], preamble_text: str) -> None
            post_callback: Function to call after tool execution.
                              Must have signature (tool_name: str, tool_input: Dict[str, Any], result: Any) -> Any
                              
        Raises:
            ValueError: If the callbacks don't have the correct signature
        """
        # Check pre_callback signature if provided
        if pre_callback:
            import inspect
            sig = inspect.signature(pre_callback)
            params = list(sig.parameters.keys())
            if len(params) != 3:
                raise ValueError(f"Pre-callback must have exactly 3 parameters: (tool_name, tool_input, preamble_text). Got {len(params)} parameters: {params}")
            if params[0] != "tool_name" or params[1] != "tool_input" or params[2] != "preamble_text":
                raise ValueError(f"Pre-callback must have parameters named 'tool_name', 'tool_input', and 'preamble_text' in that order. Got: {params}")
        
        # Check post_callback signature if provided
        if post_callback:
            import inspect
            sig = inspect.signature(post_callback)
            params = list(sig.parameters.keys())
            if len(params) != 3:
                raise ValueError(f"Post-callback must have exactly 3 parameters: (tool_name, tool_input, result). Got {len(params)} parameters: {params}")
            if params[0] != "tool_name" or params[1] != "tool_input" or params[2] != "result":
                raise ValueError(f"Post-callback must have parameters named 'tool_name', 'tool_input', and 'result' in that order. Got: {params}")
        
        self.pre_tool_callback = pre_callback
        self.post_tool_callback = post_callback
    
    def get_tool_schemas(self) -> List[Dict]:
        """
        Get JSON schemas for all registered tools.
        
        Returns:
            List of tool schemas
        """
        schemas = []
        
        for name, func in self.tools.items():
            # Special handling for text editor tool
            if name == "str_replace_editor" and self.text_editor_tool:
                # For text editor, only include name and type
                schemas.append({
                    "name": "str_replace_editor",
                    "type": "text_editor_20250124"
                })
            else:
                schema = generate_tool_schema(func, name)
                schemas.append(schema)
        
        return schemas
    
    def execute_tool(self, tool_name: str, tool_input: Dict[str, Any], preamble_text: str = "") -> Union[str, Tuple[str, bool]]:
        """
        Execute a tool with the given input.
        
        Args:
            tool_name: Name of the tool to execute
            tool_input: Input parameters for the tool
            preamble_text: Any text generated before the tool call
            
        Returns:
            Tool execution result as a string or a tuple of (content, is_error)
        """
        # Check if this is a text editor tool request
        if tool_name == "str_replace_editor" and self.text_editor_tool:
            tool_func = self.text_editor_tool
        else:
            # Get the tool function
            tool_func = self.tools.get(tool_name)
        
        if not tool_func:
            return (f"Error: Tool '{tool_name}' not found", True)
        
        # Call pre-callback if available
        if self.pre_tool_callback:
            self.pre_tool_callback(tool_name, tool_input, preamble_text)
        
        # Execute the tool
        result = tool_func(**tool_input)
        
        # Call post-callback if available
        if self.post_tool_callback:
            result = self.post_tool_callback(tool_name, tool_input, result)
        
        # Handle tuple case for error reporting
        if isinstance(result, tuple) and len(result) == 2 and isinstance(result[1], bool):
            content, is_error = result
            # Convert content to string if it's not already
            if not isinstance(content, str):
                if isinstance(content, (dict, list)):
                    content = json.dumps(content)
                else:
                    content = str(content)
            return (content, is_error)
        
        # Convert result to string if it's not already
        if not isinstance(result, str):
            if isinstance(result, (dict, list)):
                result = json.dumps(result)
            else:
                result = str(result)
        
        return result


#!/usr/bin/env python3
"""
Script to run the examples that demonstrate cost information.
"""
import os
import subprocess
import sys
from pathlib import Path

def run_example(example_path):
    """Run a single example file using uv and return its exit code."""
    print(f"\n{'=' * 80}")
    print(f"Running example: {example_path.name}")
    print(f"{'=' * 80}")
    
    result = subprocess.run(["uv", "run", str(example_path)], capture_output=False)
    
    print(f"\nExit code: {result.returncode}")
    if result.returncode != 0:
        print(f"Error running {example_path.name}. Stopping execution.")
        sys.exit(result.returncode)
    return result.returncode

def main():
    """Run the cost-related example files."""
    # Get the directory of this script
    examples_dir = Path(__file__).parent
    
    # List of examples that demonstrate cost information
    cost_examples = [
        "token_tracker.py",
        "token_usage.py",
        "token_usage_tracking.py",
        "token_usage_with_tools.py",
        "tool_token_tracking.py",
        "agent_with_token_tracking.py"
    ]
    
    # Convert to full paths
    example_files = [examples_dir / example for example in cost_examples]
    
    print(f"Found {len(example_files)} cost-related examples to run")
    
    # Run each example
    for example_file in example_files:
        if example_file.exists():
            run_example(example_file)
        else:
            print(f"Warning: Example file {example_file.name} not found")
    
    print("\n" + "=" * 80)
    print("All cost-related examples ran successfully!")
    
    return 0

if __name__ == "__main__":
    sys.exit(main())



from .agent import Agent
from .tools.manager import ToolManager
from .token_tracking import TokenTracker
from .api.client import ApiClient
from .api.models import ResponseType, ToolUseResponse, TextResponse
from .tools.callbacks import create_logging_callbacks
from .utils.helpers import generate_message_id, format_tool_result

__all__ = [
    "Agent", 
    "ToolManager", 
    "TokenTracker",
    "ApiClient",
    "ResponseType",
    "ToolUseResponse", 
    "TextResponse",
    "create_logging_callbacks",
    "generate_message_id",
    "format_tool_result"
]



#!/usr/bin/env python3
from claudine import Agent
import sys
import json
import random

def main():
    """
    Example that demonstrates how to track token usage and cost for tool-related requests.
    """
    # Define tool functions
    def get_weather(location):
        """Get the current weather for a location."""
        # Simulate weather data
        weather_data = {
            "temperature": random.randint(60, 90),
            "condition": random.choice(["Sunny", "Cloudy", "Rainy", "Partly cloudy"]),
            "humidity": random.randint(30, 80),
            "wind": f"{random.randint(5, 20)} mph"
        }
        return json.dumps(weather_data)
    
    def calculate(expression):
        """Calculate the result of a mathematical expression."""
        try:
            # Warning: eval can be dangerous in production code
            result = eval(expression, {"__builtins__": {}}, {})
            return f"Result: {result}"
        except Exception as e:
            return f"Error: {str(e)}"
    
    # Initialize Agent with tools
    agent = Agent(
        max_tokens=1000, 
        temperature=0.7,
        tools=[get_weather, calculate]
    )
    
    # First prompt that will use a tool
    print("=" * 50)
    print("FIRST PROMPT (TOOL USE)")
    print("=" * 50)
    first_prompt = "What's the weather like in San Francisco?"
    first_response = agent.process_prompt(first_prompt)
    
    print(f"User: {first_prompt}")
    print(f"Claude: {first_response}")
    
    # Get token usage information
    token_info = agent.get_token_usage()
    
    print("\nToken Usage Information:")
    print(f"Text input tokens: {token_info.text_usage.input_tokens}")
    print(f"Text output tokens: {token_info.text_usage.output_tokens}")
    print(f"Text total tokens: {token_info.text_usage.total_tokens}")
    print(f"Tool input tokens: {token_info.tools_usage.input_tokens}")
    print(f"Tool output tokens: {token_info.tools_usage.output_tokens}")
    print(f"Tool total tokens: {token_info.tools_usage.total_tokens}")
    
    # Get cost information
    cost_info = agent.get_cost()
    
    print("\nCost Information:")
    print(f"Text input cost: ${cost_info['text_cost'].input_cost:.6f} {cost_info['text_cost'].unit}")
    print(f"Text output cost: ${cost_info['text_cost'].output_cost:.6f} {cost_info['text_cost'].unit}")
    print(f"Text total cost: ${cost_info['text_cost'].total_cost:.6f} {cost_info['text_cost'].unit}")
    print(f"Tool input cost: ${cost_info['tools_cost'].input_cost:.6f} {cost_info['tools_cost'].unit}")
    print(f"Tool output cost: ${cost_info['tools_cost'].output_cost:.6f} {cost_info['tools_cost'].unit}")
    print(f"Tool total cost: ${cost_info['tools_cost'].total_cost:.6f} {cost_info['tools_cost'].unit}")
    print(f"Total cost: ${cost_info['total_cost'].total_cost:.6f} {cost_info['total_cost'].unit}")
    
    # Second prompt that will use another tool
    print("=" * 50)
    print("SECOND PROMPT (TOOL USE)")
    print("=" * 50)
    second_prompt = "Calculate 123 * 456"
    second_response = agent.process_prompt(second_prompt)
    
    print(f"User: {second_prompt}")
    print(f"Claude: {second_response}")
    
    # Get updated token usage information
    token_info = agent.get_token_usage()
    
    # Print token usage by tool
    print("\nToken Usage By Tool:")
    for tool_name, usage in token_info.by_tool.items():
        print(f"Tool: {tool_name}")
        print(f"  Input tokens: {usage.input_tokens}")
        print(f"  Output tokens: {usage.output_tokens}")
        print(f"  Total tokens: {usage.total_tokens}")
    
    # Get updated cost information
    cost_info = agent.get_cost()
    
    # Print cost by tool
    print("\nCost By Tool:")
    for tool_name, cost in cost_info['by_tool'].items():
        print(f"Tool: {tool_name}")
        print(f"  Input cost: ${cost.input_cost:.6f} {cost.unit}")
        print(f"  Output cost: ${cost.output_cost:.6f} {cost.unit}")
        print(f"  Total cost: ${cost.total_cost:.6f} {cost.unit}")
    
    # Print total token usage and cost
    print("\nTotal Token Usage and Cost:")
    print(f"Total input tokens: {token_info.total_usage.input_tokens}")
    print(f"Total output tokens: {token_info.total_usage.output_tokens}")
    print(f"Total tokens: {token_info.total_usage.total_tokens}")
    print(f"Total cost: ${cost_info['total_cost'].total_cost:.6f} {cost_info['total_cost'].unit}")
    
    return 0

if __name__ == "__main__":
    sys.exit(main())



"""
Custom exceptions for the claudine library.
"""

class ClaudineException(Exception):
    """Base exception for all claudine exceptions."""
    pass

class MaxTokensExceededException(ClaudineException):
    """
    Exception raised when the response from Claude is truncated 
    because it reached the maximum token limit.
    """
    def __init__(self, message="Response was truncated because it reached the maximum token limit", 
                 response_text=None):
        self.message = message
        self.response_text = response_text
        super().__init__(self.message)

class MaxRoundsExceededException(ClaudineException):
    """
    Exception raised when the maximum number of tool execution rounds is reached.
    """
    def __init__(self, message="Maximum number of tool execution rounds reached. Some tasks may be incomplete", 
                 response_text=None, rounds=None):
        self.message = message
        self.response_text = response_text
        self.rounds = rounds
        super().__init__(self.message)



"""
Tool callback functionality for Claude.
"""
from typing import Dict, Any, Callable, Optional

# Type definitions for callbacks
PreCallbackType = Callable[[str, Dict[str, Any], str], None]
PostCallbackType = Callable[[str, Dict[str, Any], Any], Any]

def create_logging_callbacks(log_prefix: str = "Tool"):
    """
    Create simple logging callbacks for tool execution.
    
    Args:
        log_prefix: Prefix for log messages
        
    Returns:
        Tuple of (pre_callback, post_callback)
    """
    def pre_callback(tool_name: str, tool_input: Dict[str, Any], preamble_text: str) -> None:
        """Log before tool execution."""
        print(f"{log_prefix} Executing: {tool_name}")
        print(f"{log_prefix} Input: {tool_input}")
        if preamble_text:
            print(f"{log_prefix} Preamble: {preamble_text[:100]}...")
    
    def post_callback(tool_name: str, tool_input: Dict[str, Any], result: Any) -> Any:
        """Log after tool execution."""
        print(f"{log_prefix} Result: {result}")
        return result
    
    return pre_callback, post_callback



#!/usr/bin/env python
"""
Run examples with UTF-8 encoding to properly display Unicode characters like the cent symbol (¢).
Usage: python run_example.py <example_module_name>
Example: python run_example.py token_cost
"""

import sys
import importlib
import os

def main():
    if len(sys.argv) < 2:
        print("Usage: python run_example.py <example_module_name>")
        print("Example: python run_example.py token_cost")
        return 1
    
    # Set UTF-8 encoding for the terminal
    if sys.platform == "win32":
        os.system("chcp 65001 > nul")
    
    # Import and run the specified example
    example_name = sys.argv[1]
    try:
        module = importlib.import_module(f"claudine.examples.{example_name}")
        return module.main()
    except ImportError:
        print(f"Error: Could not import example module '{example_name}'")
        print("Available examples:")
        for file in os.listdir(os.path.dirname(__file__)):
            if file.endswith(".py") and file != "__init__.py" and file != "run_example.py":
                print(f"  {file[:-3]}")
        return 1

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
"""
Script to run all example files in the claudine project.
"""
import os
import subprocess
import sys
from pathlib import Path

def run_example(example_path):
    """Run a single example file using uv and return its exit code."""
    print(f"\n{'=' * 80}")
    print(f"Running example: {example_path.name}")
    print(f"{'=' * 80}")
    
    result = subprocess.run(["uv", "run", str(example_path)], capture_output=False)
    
    print(f"\nExit code: {result.returncode}")
    if result.returncode != 0:
        print(f"Error running {example_path.name}. Stopping execution.")
        sys.exit(result.returncode)
    return result.returncode

def main():
    """Run all example files in the examples directory."""
    # Get the directory of this script
    examples_dir = Path(__file__).parent
    
    # Check if a specific example was specified
    if len(sys.argv) > 1:
        example_name = sys.argv[1]
        example_path = examples_dir / example_name
        
        # Check if the file exists
        if not example_path.exists():
            print(f"Error: Example file '{example_name}' not found")
            return 1
        
        # Run the example
        return run_example(example_path)
    
    # Find all Python files in the examples directory
    example_files = [
        f for f in examples_dir.glob("*.py") 
        if f.name not in ["run_all_examples.py", "run_cost_examples.py"]
    ]
    
    # Sort files to ensure consistent order
    example_files.sort()
    
    print(f"Found {len(example_files)} example files to run")
    
    # Run each example
    for example_file in example_files:
        run_example(example_file)
    
    print("\n" + "=" * 80)
    print("All examples ran successfully!")
    
    return 0

if __name__ == "__main__":
    sys.exit(main())



#!/usr/bin/env python3
from claudine import Agent
import sys
import os
from datetime import datetime

def main():
    """
    Simple command-line application that demonstrates the Claudine Agent.
    """
   
    # Define tools
    tools = [get_time, get_weather, search_info, calculate]
    
    # Define instructions for the agent
    instructions = """
    You are a helpful AI assistant named Claude. Your primary goal is to assist the user with 
    information, calculations, and general questions. When using tools, be precise and efficient.
    Always explain your reasoning clearly and concisely.
    """
    
    # Define tool callbacks
    def custom_pre_callback(tool_name, tool_input, preamble_text):
        print(f" About to execute: {tool_name}")
        print(f" Input parameters: {tool_input}")
        print(f" Preamble text: '{preamble_text if preamble_text else ''}'")
        
        # You can modify the tool input here if needed
        # For example, add defaults or transform inputs
        return tool_input
    
    def custom_post_callback(tool_name, tool_input, result):
        print(f" Tool executed successfully: {tool_name}")
        print(f" Result: {result}")
        
        # You can modify the result here if needed
        # For example, format it differently or add metadata
        return result
    
    # Initialize Agent with model parameters, tools, and callbacks
    agent = Agent(
        instructions=instructions,
        tools=tools,
        tool_callbacks=(custom_pre_callback, custom_post_callback)
    )
    
    # Process user prompts
    prompts = [
        "What time is it now, and what's the weather in San Francisco?",
        "Can you calculate 25 * 18 for me?"
    ]
    
    for prompt in prompts:
        print(f"User: {prompt}")
        response = agent.process_prompt(prompt)
        print(f"Claude: {response}")
        print()
        
    # Reset conversation history
    agent.reset()
    
    return 0

# Define tool functions
def get_time():
    """Returns the current date and time."""
    now = datetime.now()
    return now.strftime("%Y-%m-%d %H:%M:%S")

def get_weather(location: str):
    """
    Get the current weather in a given location.
    
    Args:
        location: The city and state, e.g. San Francisco, CA
    """
    # Simulate weather API call
    return f"72°F and sunny in {location}"

def search_info(query: str):
    """
    Search for information on a given topic.
    
    Args:
        query: The search query
    """
    # Simulate search results
    return f"Here are search results for '{query}':\n- Result 1\n- Result 2\n- Result 3"

def calculate(expression: str):
    """
    Evaluate a mathematical expression.
    
    Args:
        expression: A mathematical expression as a string, e.g. "2 * (3 + 4)"
    """
    try:
        # Note: In production, you should use a safer evaluation method
        result = eval(expression)
        return f"Result: {result}"
    except Exception as e:
        return f"Error calculating: {str(e)}"

if __name__ == "__main__":
    sys.exit(main())



"""
Example demonstrating the use of the text_editor parameter in the Agent class.
"""
import os
from claudine.agent import Agent

def handle_editor_tool(**kwargs):
    """
    Handle text editor tool requests from Claude.
    Simply prints the command name and returns the kwargs as a string.
    
    Args:
        **kwargs: All arguments passed to the tool
        
    Returns:
        String representation of the arguments
    """
    cmd_name = kwargs.get("command")
    print(f"Text editor command: {cmd_name}")
    print(f"All arguments: {kwargs}")
    
    # Return a string representation of the arguments
    return str(kwargs)

def main():
    # Get API key from environment variable
    api_key = os.environ.get("ANTHROPIC_API_KEY")
    
    # Initialize the agent with the text editor handler
    agent = Agent(
        api_key=api_key,
        max_tokens=1024,
        temperature=0.7,
        text_editor_tool=handle_editor_tool,
        debug_mode=True  # Enable debug mode to see messages sent to Claude
    )
    
    # Example prompt that might trigger the text editor tool
    response = agent.process_prompt(
        "I have an error in file.py. The function add(a, b) is subtracting instead of adding. "
        "Can you check the file and fix it?"
    )
    
    # Get and print token usage and cost information
    token_usage = agent.get_token_usage()
    cost_info = agent.get_cost()
    
    print("\nToken Usage and Cost Information:")
    print(f"Total cost: {cost_info.format_total_cost()}")
    print(f"Total input tokens: {token_usage.total_usage.input_tokens}")
    print(f"Total output tokens: {token_usage.total_usage.output_tokens}")
    print(f"Text input tokens: {token_usage.text_usage.input_tokens}")
    print(f"Text output tokens: {token_usage.text_usage.output_tokens}")
    print(f"Tools input tokens: {token_usage.tools_usage.input_tokens}")
    print(f"Tools output tokens: {token_usage.tools_usage.output_tokens}")
    
    print("\nClaude's response:")
    print(response)

if __name__ == "__main__":
    main()



"""
Tool schema generation for Claude.
"""
from typing import Dict, List, Optional, Callable, Any, Tuple
import inspect
import json

def generate_tool_schema(func: Callable, name: Optional[str] = None) -> Dict:
    """
    Generate a JSON schema for a tool function.
    
    Args:
        func: Function to generate schema for
        name: Optional name for the tool (defaults to function name)
        
    Returns:
        Tool schema as a dictionary
    """
    # Get function signature
    sig = inspect.signature(func)
    
    # Get function name and docstring
    func_name = name if name else func.__name__
    func_desc = func.__doc__ or f"Function {func_name}"
    
    # Create parameter schema
    properties = {}
    required = []
    
    for param_name, param in sig.parameters.items():
        # Skip self parameter for methods
        if param_name == "self":
            continue
        
        # Add parameter to properties
        properties[param_name] = {
            "type": "string",  # Default to string type
            "description": f"Parameter {param_name}"
        }
        
        # Add to required list if no default value
        if param.default == inspect.Parameter.empty:
            required.append(param_name)
    
    # Create schema
    schema = {
        "name": func_name,
        "description": func_desc,
        "input_schema": {
            "type": "object",
            "properties": properties,
            "required": required
        }
    }
    
    return schema

def validate_tool_input(schema: Dict, input_data: Dict) -> Tuple[bool, Optional[str]]:
    """
    Validate tool input against its schema.
    
    Args:
        schema: Tool schema
        input_data: Input data to validate
        
    Returns:
        Tuple of (is_valid, error_message)
    """
    # Get required fields
    required = schema["input_schema"].get("required", [])
    
    # Check required fields
    for field in required:
        if field not in input_data:
            return False, f"Missing required field: {field}"
    
    # Validation passed
    return True, None



"""
Example demonstrating the callback signature validation.
"""
from claudine import Agent, create_logging_callbacks

def main():
    # These callbacks have the correct signature
    pre_callback, post_callback = create_logging_callbacks()
    
    # This will work fine
    agent = Agent(
        tool_callbacks=(pre_callback, post_callback)
    )
    print("Successfully set up agent with valid callbacks")
    
    # Invalid pre-callback (wrong number of parameters)
    def invalid_pre_callback(tool_name):
        print(f"Invalid pre-callback called for: {tool_name}")
    
    try:
        agent = Agent(
            tool_callbacks=(invalid_pre_callback, post_callback)
        )
        print("This should not be printed")
    except ValueError as e:
        print(f"Caught expected error: {e}")
    
    # Invalid post-callback (wrong number of parameters)
    def invalid_post_callback(tool_name, tool_input):
        print(f"Invalid post-callback called for: {tool_name}")
        return "Modified result"
    
    try:
        agent = Agent(
            tool_callbacks=(pre_callback, invalid_post_callback)
        )
        print("This should not be printed")
    except ValueError as e:
        print(f"Caught expected error: {e}")
    
    print("All tests completed")

if __name__ == "__main__":
    main()



from claudine.api.models import TokenUsage, TokenUsageInfo
from claudine.token_tracking import TokenTracker

# Create a token tracker and add some test messages
tracker = TokenTracker()
tracker.add_message('msg1', 100, 200, is_tool_related=False)
tracker.add_message('msg2', 150, 250, is_tool_related=True, tool_name='calculator')

# Get the token usage information
usage_info = tracker.get_token_usage()

# Print the results
print("=== Token Usage Test ===")
print(f"Text usage: {usage_info.text_usage.input_tokens} input, {usage_info.text_usage.output_tokens} output, {usage_info.text_usage.total_tokens} total")
print(f"Tools usage: {usage_info.tools_usage.input_tokens} input, {usage_info.tools_usage.output_tokens} output, {usage_info.tools_usage.total_tokens} total")
print(f"Total usage: {usage_info.total_usage.input_tokens} input, {usage_info.total_usage.output_tokens} output, {usage_info.total_usage.total_tokens} total")

print("\n=== By Tool ===")
for tool_name, usage in usage_info.by_tool.items():
    print(f"Tool: {tool_name}")
    print(f"  Input tokens: {usage.input_tokens}")
    print(f"  Output tokens: {usage.output_tokens}")
    print(f"  Total tokens: {usage.total_tokens}")



from dataclasses import dataclass
from typing import Dict, Optional, Union

@dataclass
class TokenUsage:
    """Represents token usage information."""
    input_tokens: int = 0
    output_tokens: int = 0
    
    @property
    def total_tokens(self) -> int:
        """Calculate total tokens from input and output tokens"""
        return self.input_tokens + self.output_tokens

@dataclass
class TokenUsageInfo:
    """Comprehensive token usage information including text and tool usage."""
    text_usage: TokenUsage
    tools_usage: TokenUsage
    by_tool: Dict[str, TokenUsage] = None
    
    @property
    def total_usage(self) -> TokenUsage:
        """Get combined total usage across text and tools"""
        return TokenUsage(
            input_tokens=self.text_usage.input_tokens + self.tools_usage.input_tokens,
            output_tokens=self.text_usage.output_tokens + self.tools_usage.output_tokens
        )

class TokenTracker:
    """Tracks token usage for Claude API calls."""
    
    def __init__(self):
        """Initialize an empty token tracker."""
        self.messages = {}  # Dictionary to store message token usage by message ID
    
    def add_message(self, message_id: str, input_tokens: int, output_tokens: int, 
                   is_tool_related: bool = False, tool_name: Optional[str] = None,
                   parent_message_id: Optional[str] = None):
        """Add a message's token usage to the tracker."""
        self.messages[message_id] = {
            "message_id": message_id,
            "input_tokens": input_tokens,
            "output_tokens": output_tokens,
            "is_tool_related": is_tool_related,
            "tool_name": tool_name,
            "parent_message_id": parent_message_id
        }
    
    def get_token_usage(self, message_id: Optional[str] = None) -> Union[Dict, TokenUsageInfo]:
        """Get token usage information for a specific message or consolidated usage."""
        if message_id:
            return self.messages.get(message_id, {})
        
        # Calculate total usage
        total_input = sum(msg["input_tokens"] for msg in self.messages.values())
        total_output = sum(msg["output_tokens"] for msg in self.messages.values())
        
        # Calculate tool-related usage
        tool_messages = [msg for msg in self.messages.values() if msg["is_tool_related"]]
        tool_input = sum(msg["input_tokens"] for msg in tool_messages)
        tool_output = sum(msg["output_tokens"] for msg in tool_messages)
        
        # Calculate text-only usage
        text_input = total_input - tool_input
        text_output = total_output - tool_output
        
        # Create TokenUsage objects
        text_usage = TokenUsage(
            input_tokens=text_input,
            output_tokens=text_output
        )
        
        tools_usage = TokenUsage(
            input_tokens=tool_input,
            output_tokens=tool_output
        )
        
        # Calculate usage by tool
        by_tool = {}
        for msg in self.messages.values():
            if msg["is_tool_related"] and msg["tool_name"]:
                tool_name = msg["tool_name"]
                
                if tool_name not in by_tool:
                    by_tool[tool_name] = {
                        "input_tokens": 0,
                        "output_tokens": 0
                    }
                
                by_tool[tool_name]["input_tokens"] += msg["input_tokens"]
                by_tool[tool_name]["output_tokens"] += msg["output_tokens"]
        
        # Convert by_tool to use TokenUsage objects
        by_tool_usage = {}
        for tool_name, usage in by_tool.items():
            by_tool_usage[tool_name] = TokenUsage(
                input_tokens=usage["input_tokens"],
                output_tokens=usage["output_tokens"]
            )
        
        return TokenUsageInfo(
            text_usage=text_usage,
            tools_usage=tools_usage,
            by_tool=by_tool_usage
        )

# Test the implementation
if __name__ == "__main__":
    # Create a token tracker and add some test messages
    tracker = TokenTracker()
    tracker.add_message('msg1', 100, 200, is_tool_related=False)
    tracker.add_message('msg2', 150, 250, is_tool_related=True, tool_name='calculator')
    tracker.add_message('msg3', 75, 125, is_tool_related=True, tool_name='weather')
    
    # Get the token usage information
    usage_info = tracker.get_token_usage()
    
    # Print the results
    print("=== Token Usage Test ===")
    print(f"Text usage: {usage_info.text_usage.input_tokens} input, {usage_info.text_usage.output_tokens} output, {usage_info.text_usage.total_tokens} total")
    print(f"Tools usage: {usage_info.tools_usage.input_tokens} input, {usage_info.tools_usage.output_tokens} output, {usage_info.tools_usage.total_tokens} total")
    print(f"Total usage: {usage_info.total_usage.input_tokens} input, {usage_info.total_usage.output_tokens} output, {usage_info.total_usage.total_tokens} total")
    
    print("\n=== By Tool ===")
    for tool_name, usage in usage_info.by_tool.items():
        print(f"Tool: {tool_name}")
        print(f"  Input tokens: {usage.input_tokens}")
        print(f"  Output tokens: {usage.output_tokens}")
        print(f"  Total tokens: {usage.total_tokens}")



#!/usr/bin/env python3
from claudine import Agent
import sys

def main():
    """
    Example that demonstrates how to track token usage and cost with the Claudine Agent.
    """
    # Initialize Agent
    agent = Agent(max_tokens=100, temperature=0.7)
    
    # Make a simple API call
    print("Making API call...")
    response = agent.process_prompt("What is token counting and why is it important?")
    
    # Print the response
    print("\nClaude's response:")
    print(response)
    
    # Get token usage information
    token_info = agent.get_token_usage()
    
    # Print token usage information
    print("\nToken Usage Information:")
    print(f"Input tokens: {token_info.text_usage.input_tokens}")
    print(f"Output tokens: {token_info.text_usage.output_tokens}")
    print(f"Total tokens: {token_info.text_usage.total_tokens}")
    
    # Get cost information
    cost_info = agent.get_cost()
    
    print("\nCost Information:")
    print(f"Input cost: {cost_info.format_input_cost()}")
    print(f"Output cost: {cost_info.format_output_cost()}")
    print(f"Total cost: {cost_info.format_total_cost()}")
    
    return 0

if __name__ == "__main__":
    sys.exit(main())



"""
Data models for API requests and responses.
"""
from typing import Dict, List, Optional, Any, Union
from dataclasses import dataclass

@dataclass
class ToolUseResponse:
    """
    Represents a tool use response from Claude.
    """
    type: str = "tool_use"
    name: str = ""
    input: Dict[str, Any] = None
    id: str = ""
    message_id: str = ""
    preamble: str = ""

@dataclass
class TextResponse:
    """
    Represents a text response from Claude.
    """
    type: str = "text"
    text: str = ""
    message_id: str = ""
    was_truncated: bool = False

@dataclass
class TokenPricing:
    """
    Represents pricing information for tokens.
    """
    cost_per_million_tokens: float = 0.0
    unit: str = "USD"
    
    def calculate_cost(self, tokens: int) -> float:
        """Calculate cost for a given number of tokens"""
        return (tokens / 1_000_000) * self.cost_per_million_tokens

@dataclass
class ModelPricing:
    """
    Represents pricing information for a model.
    """
    input_tokens: TokenPricing
    output_tokens: TokenPricing
    
    def calculate_cost(self, input_tokens: int, output_tokens: int) -> float:
        """Calculate total cost for input and output tokens"""
        input_cost = self.input_tokens.calculate_cost(input_tokens)
        output_cost = self.output_tokens.calculate_cost(output_tokens)
        return input_cost + output_cost

@dataclass
class TokenUsage:
    """
    Represents token usage information.
    """
    input_tokens: int = 0
    output_tokens: int = 0
    
    @property
    def total_tokens(self) -> int:
        """Calculate total tokens from input and output tokens"""
        return self.input_tokens + self.output_tokens
    
    def calculate_cost(self, pricing: ModelPricing) -> float:
        """Calculate cost based on token usage and pricing"""
        return pricing.calculate_cost(self.input_tokens, self.output_tokens)

@dataclass
class TokenCost:
    """
    Represents cost information for token usage.
    """
    input_cost: float = 0.0
    output_cost: float = 0.0
    unit: str = "USD"
    
    @property
    def total_cost(self) -> float:
        """Calculate total cost from input and output costs"""
        return self.input_cost + self.output_cost
    
    def format_input_cost(self) -> str:
        """Format input cost in dollars or cents based on value"""
        # Unicode cent symbol (requires UTF-8 terminal)
        if self.input_cost < 1.0:
            return f"{self.input_cost * 100:.2f}¢"
        return f"${self.input_cost:.6f}"
    
    def format_output_cost(self) -> str:
        """Format output cost in dollars or cents based on value"""
        # Unicode cent symbol (requires UTF-8 terminal)
        if self.output_cost < 1.0:
            return f"{self.output_cost * 100:.2f}¢"
        return f"${self.output_cost:.6f}"
    
    def format_total_cost(self) -> str:
        """Format total cost in dollars or cents based on value"""
        # Unicode cent symbol (requires UTF-8 terminal)
        if self.total_cost < 1.0:
            return f"{self.total_cost * 100:.2f}¢"
        return f"${self.total_cost:.6f}"

@dataclass
class TokenUsageInfo:
    """
    Comprehensive token usage information including text and tool usage.
    """
    text_usage: TokenUsage
    tools_usage: TokenUsage
    by_tool: Dict[str, TokenUsage] = None
    
    @property
    def total_usage(self) -> TokenUsage:
        """Get combined total usage across text and tools"""
        return TokenUsage(
            input_tokens=self.text_usage.input_tokens + self.tools_usage.input_tokens,
            output_tokens=self.text_usage.output_tokens + self.tools_usage.output_tokens
        )
    
    def calculate_cost(self, pricing: ModelPricing) -> Dict[str, TokenCost]:
        """
        Calculate costs for all token usage based on provided pricing.
        
        Returns:
            Dictionary with text_cost, tools_cost, total_cost and by_tool costs
        """
        # Calculate text costs
        text_input_cost = pricing.input_tokens.calculate_cost(self.text_usage.input_tokens)
        text_output_cost = pricing.output_tokens.calculate_cost(self.text_usage.output_tokens)
        text_cost = TokenCost(
            input_cost=text_input_cost,
            output_cost=text_output_cost,
            unit=pricing.input_tokens.unit
        )
        
        # Calculate tools costs
        tools_input_cost = pricing.input_tokens.calculate_cost(self.tools_usage.input_tokens)
        tools_output_cost = pricing.output_tokens.calculate_cost(self.tools_usage.output_tokens)
        tools_cost = TokenCost(
            input_cost=tools_input_cost,
            output_cost=tools_output_cost,
            unit=pricing.input_tokens.unit
        )
        
        # Calculate total cost
        total_cost = TokenCost(
            input_cost=text_input_cost + tools_input_cost,
            output_cost=text_output_cost + tools_output_cost,
            unit=pricing.input_tokens.unit
        )
        
        # Calculate by_tool costs
        by_tool_cost = {}
        if self.by_tool:
            for tool_name, usage in self.by_tool.items():
                tool_input_cost = pricing.input_tokens.calculate_cost(usage.input_tokens)
                tool_output_cost = pricing.output_tokens.calculate_cost(usage.output_tokens)
                by_tool_cost[tool_name] = TokenCost(
                    input_cost=tool_input_cost,
                    output_cost=tool_output_cost,
                    unit=pricing.input_tokens.unit
                )
        
        return {
            "text_cost": text_cost,
            "tools_cost": tools_cost,
            "total_cost": total_cost,
            "by_tool": by_tool_cost
        }

ResponseType = Union[ToolUseResponse, TextResponse]



#!/usr/bin/env python3
from claudine import Agent
import sys
import json

def weather_tool(location: str, unit: str = "celsius"):
    """
    Get the current weather for a location.
    
    Args:
        location: The location to get weather for
        unit: Temperature unit (celsius or fahrenheit)
        
    Returns:
        Weather information as a string
    """
    # This is a mock implementation
    weather_data = {
        "location": location,
        "temperature": 22 if unit == "celsius" else 72,
        "conditions": "Sunny",
        "humidity": "45%"
    }
    
    return json.dumps(weather_data)

def calculator(expression: str):
    """
    Calculate the result of a mathematical expression.
    
    Args:
        expression: Mathematical expression to evaluate
        
    Returns:
        Result of the calculation
    """
    try:
        # Warning: eval can be dangerous in production code
        # This is just for demonstration purposes
        result = eval(expression)
        return str(result)
    except Exception as e:
        return f"Error: {str(e)}"

def main():
    """
    Example that demonstrates token usage tracking with tools.
    """
    
    # Initialize Agent with tools
    agent = Agent(
        max_tokens=1000, 
        temperature=0.7,
        tools=[weather_tool, calculator]
    )
    
    # Process a prompt that will likely use tools
    response = agent.process_prompt(
        "What's the weather in New York? Also, what is 1234 * 5678?"
    )
    
    # Print the response
    print("Claude's response:")
    print(response)
    print("\n" + "-" * 50 + "\n")
    
    # Get token usage information
    token_info = agent.get_token_usage()
    
    print("Token Usage Information:")
    print(f"Text input tokens: {token_info.text_usage.input_tokens}")
    print(f"Text output tokens: {token_info.text_usage.output_tokens}")
    print(f"Text total tokens: {token_info.text_usage.total_tokens}")
    print(f"Tool input tokens: {token_info.tools_usage.input_tokens}")
    print(f"Tool output tokens: {token_info.tools_usage.output_tokens}")
    print(f"Tool total tokens: {token_info.tools_usage.total_tokens}")
    
    # Print total token usage
    print("\n" + "-" * 50 + "\n")
    print("Total Token Usage:")
    total = token_info.total_usage
    print(f"Total input tokens: {total.input_tokens}")
    print(f"Total output tokens: {total.output_tokens}")
    print(f"Total tokens: {total.total_tokens}")
    
    # Get cost information
    cost_info = agent.get_cost()
    
    print("\n" + "-" * 50 + "\n")
    print("Cost Information:")
    print(f"Input cost: {cost_info.format_input_cost()} {cost_info.unit}")
    print(f"Output cost: {cost_info.format_output_cost()} {cost_info.unit}")
    print(f"Total cost: {cost_info.format_total_cost()} {cost_info.unit}")
    
    # Print token usage by tool
    print("\n" + "-" * 50 + "\n")
    print("Token Usage by Tool:")
    for tool_name, usage in token_info.by_tool.items():
        print(f"\nTool: {tool_name}")
        print(f"  Input tokens: {usage.input_tokens}")
        print(f"  Output tokens: {usage.output_tokens}")
        print(f"  Total tokens: {usage.total_tokens}")
        
    return 0

if __name__ == "__main__":
    sys.exit(main())



"""
Main Agent class for Claudine.
Provides a high-level interface for interacting with Claude models,
with support for tool use, token tracking, and conversation management.
"""
from typing import Dict, List, Optional, Callable, Any, Tuple, Union
import anthropic

from .api.client import ApiClient
from .api.models import ResponseType, ToolUseResponse, TextResponse, TokenUsage, TokenUsageInfo
from .tools.manager import ToolManager
from .token_tracking import TokenTracker, DEFAULT_MODEL
from .utils.helpers import generate_message_id, extract_text_content, format_tool_result
from .exceptions import MaxTokensExceededException, MaxRoundsExceededException

class Agent:
    """
    Agent for interacting with Claude.
    """
    
    def __init__(self, api_key: Optional[str] = None, 
                max_tokens: int = 1024, temperature: float = 0.7,
                max_rounds: int = 30, instructions: Optional[str] = None,
                tools: Optional[List[Callable]] = None,
                tool_callbacks: Optional[Tuple[Optional[Callable], Optional[Callable]]] = None,
                disable_parallel_tool_use: bool = True,
                text_editor_tool: Optional[Callable] = None,
                debug_mode: bool = False):
        """
        Initialize the Agent wrapper with your Anthropic API key, model parameters, and tools.
        If api_key is not provided, it will use the ANTHROPIC_API_KEY environment variable.
        
        Args:
            api_key: Anthropic API key
            max_tokens: Maximum number of tokens to generate
            temperature: Temperature for generation
            max_rounds: Maximum number of rounds for tool use
            instructions: Instructions to guide the model's behavior (used as system prompt)
            tools: List of functions to register as tools
            tool_callbacks: Tuple of (pre_callback, post_callback) callables for tool execution
            disable_parallel_tool_use: Disable parallel tool use to ensure accurate token accounting
            text_editor_tool: Callable function to handle text editor tool requests. Must implement the
                         commands and response formats as specified in the Anthropic documentation:
                         https://docs.anthropic.com/en/docs/build-with-claude/tool-use/text-editor-tool
            debug_mode: If True, print debug information about API calls
        """
        # Initialize API client
        self.api_client = ApiClient(api_key=api_key, debug_mode=debug_mode)
        
        # Store parameters
        self.messages = []
        self.max_tokens = max_tokens
        self.temperature = temperature
        self.max_rounds = max_rounds
        self.system = instructions
        
        # Initialize token tracker
        self.token_tracker = TokenTracker()
        
        # Initialize tool manager
        self.tool_manager = ToolManager()
        
        # Register tools if provided
        if tools:
            self.tool_manager.register_tools(tools)
        
        # Register text editor tool if provided
        if text_editor_tool:
            self.tool_manager.tools["str_replace_editor"] = text_editor_tool
            self.tool_manager.text_editor_tool = text_editor_tool
        
        # Set tool callbacks if provided
        if tool_callbacks:
            pre_callback, post_callback = tool_callbacks
            self.tool_manager.set_tool_callbacks(pre_callback, post_callback)
        
        # Disable parallel tool use to ensure accurate token accounting
        self.disable_parallel_tool_use = disable_parallel_tool_use
    
    def set_tool_callbacks(self, pre_callback: Optional[Callable] = None, 
                            post_callback: Optional[Callable] = None):
        """
        Set callbacks for tool execution.
        
        Args:
            pre_callback: Function to call before tool execution
            post_callback: Function to call after tool execution
        """
        self.tool_manager.set_tool_callbacks(pre_callback, post_callback)
    
    def _call_claude(self, tools: List[Dict]) -> ResponseType:
        """
        Call Claude with the current messages and tools.
        
        Args:
            tools: List of tool schemas
            
        Returns:
            Claude's response as a ResponseType
            
        Raises:
            MaxTokensExceededException: If the response was truncated due to token limits
        """
        # Set tool_choice with disable_parallel_tool_use parameter
        tool_choice = None
        if tools:
            tool_choice = {
                "type": "auto",
                "disable_parallel_tool_use": self.disable_parallel_tool_use
            }
        
        # Make the API call
        response = self.api_client.create_message(
            model=DEFAULT_MODEL,
            messages=self.messages,
            max_tokens=self.max_tokens,
            temperature=self.temperature,
            system=self.system,
            tools=tools,
            tool_choice=tool_choice
        )
        
        # Track token usage
        message_id = response.id
        
        # Check if this is a tool-related message
        is_tool_related = False
        tool_name = None
        parent_message_id = None
        
        # If this is a response to a tool result, it's tool-related
        if len(self.messages) >= 2 and isinstance(self.messages[-1].get("content"), list):
            for content_item in self.messages[-1].get("content", []):
                if isinstance(content_item, dict) and content_item.get("type") == "tool_result":
                    is_tool_related = True
                    # Try to find the parent message that initiated this tool call
                    if len(self.messages) >= 3:
                        for content_item in self.messages[-2].get("content", []):
                            if isinstance(content_item, dict) and content_item.get("type") == "tool_use":
                                tool_name = content_item.get("name")
                                # Find the original message that triggered this tool
                                for i in range(len(self.messages) - 3, -1, -1):
                                    if self.messages[i].get("role") == "assistant":
                                        parent_message_id = f"msg_{i}"  # Create a pseudo-ID
                                        break
        
        self.token_tracker.add_message(
            message_id=message_id,
            input_tokens=response.usage.input_tokens,
            output_tokens=response.usage.output_tokens,
            is_tool_related=is_tool_related,
            tool_name=tool_name,
            parent_message_id=parent_message_id
        )
        
        # Check if token limit was reached
        was_truncated = response.stop_reason == "max_tokens"
        
        # Check if tool use is requested
        if response.stop_reason == "tool_use":
            # Extract text and tool use from response
            text_content = ""
            tool_use = None
            
            for content_block in response.content:
                if content_block.type == "text":
                    text_content += content_block.text
                elif content_block.type == "tool_use":
                    tool_use = content_block
            
            if tool_use:
                return ToolUseResponse(
                    type="tool_use",
                    name=tool_use.name,
                    input=tool_use.input,
                    id=tool_use.id,
                    message_id=message_id,
                    preamble=text_content.strip() if text_content else None
                )
        
        # Regular text response
        text_content = extract_text_content(response.content)
        
        # If the response was truncated due to token limits, raise an exception
        if was_truncated:
            raise MaxTokensExceededException(response_text=text_content)
        
        return TextResponse(
            type="text",
            text=text_content,
            message_id=message_id,
            was_truncated=was_truncated
        )
    
    def process_prompt(self, prompt: str) -> str:
        """
        Process a prompt and return Claude's response.
        
        Args:
            prompt: User prompt
            
        Returns:
            Claude's response as a string
            
        Raises:
            MaxTokensExceededException: If the response was truncated due to token limits
            MaxRoundsExceededException: If the maximum number of tool execution rounds was reached
        """
        # Add user message to conversation
        self.messages.append({
            "role": "user",
            "content": prompt
        })
        
        # Get tool schemas
        tools = self.tool_manager.get_tool_schemas()
        
        # Call Claude
        response = self._call_claude(tools)
        
        # If tool use is requested, execute the tool
        rounds = 0
        while response.type == "tool_use" and rounds < self.max_rounds:
            # Add assistant message with tool use
            content_items = []
            if response.preamble:
                content_items.append({
                    "type": "text",
                    "text": response.preamble
                })
            
            content_items.append({
                "type": "tool_use",
                "id": response.id,
                "name": response.name,
                "input": response.input
            })
            
            self.messages.append({
                "role": "assistant",
                "content": content_items
            })
            
            # Execute the tool
            tool_result = self.tool_manager.execute_tool(response.name, response.input, response.preamble)
            
            # Add user message with tool result
            self.messages.append({
                "role": "user",
                "content": [
                    format_tool_result(response.id, tool_result)
                ]
            })
            
            # Call Claude again
            response = self._call_claude(tools)
            
            # Increment rounds
            rounds += 1
        
        # Check if we hit the max rounds limit
        if rounds >= self.max_rounds:
            raise MaxRoundsExceededException(response_text=response.text, rounds=rounds)
        
        # Add assistant message to conversation
        self.messages.append({
            "role": "assistant",
            "content": response.text
        })
        
        # Return the response text
        return response.text
    
    def get_token_usage(self) -> TokenUsageInfo:
        """
        Get token usage information.
        
        Returns:
            TokenUsageInfo object with usage details
        """
        return self.token_tracker.get_token_usage()
    
    def get_cost(self) -> Dict:
        """
        Get cost information for token usage.
        
        Returns:
            Dictionary with cost information
        """
        return self.token_tracker.get_cost()
    
    def set_model(self, model: str):
        """
        Set the model to use for the agent and update token tracker.
        
        Args:
            model: The model name
        """
        self.model = model
        self.token_tracker.set_model(model)
    
    def reset(self):
        """Reset the conversation history."""
        self.messages = []
        self.token_tracker.reset()
        
    def get_messages(self, filter_out_tools: bool = False) -> List[Dict]:
        """
        Get the current conversation messages.
        
        Args:
            filter_out_tools: If True, removes messages related to tool calls or tool results
        
        Returns:
            A copy of the conversation messages list
        """
        messages = self.messages.copy()
        
        if filter_out_tools:
            filtered_messages = []
            for message in messages:
                # Check if this is a tool-related message
                if isinstance(message.get("content"), list):
                    is_tool_message = False
                    for content_item in message["content"]:
                        if isinstance(content_item, dict) and content_item.get("type") in ["tool_use", "tool_result"]:
                            is_tool_message = True
                            break
                    
                    if is_tool_message:
                        continue
                
                filtered_messages.append(message)
            return filtered_messages
        
        return messages
    
    def set_messages(self, messages: List[Dict]):
        """
        Set the conversation messages.
        
        Args:
            messages: List of message dictionaries to set as the conversation history
        """
        self.messages = messages.copy()



from claudine import Agent

# Initialize the agent
agent = Agent()

# Process a prompt
response = agent.process_prompt("Write a short poem about programming.")
print(response)

# Get token usage information
token_info = agent.get_token_usage()
print(f"Total tokens used: {token_info.total_usage.total_tokens}")